import re
from enum import Enum

import urllib3
import semver
from lastversion import lastversion
from mutapath import Path
from remotezip import RemoteZip, RemoteIOError
from urllib3.exceptions import MaxRetryError

from kivysome.iconfonts import register, create_fontdict_file

_VERSION_MATCHER = re.compile(r'.*"version":"([a-zA-Z0-9.]+)"')
LATEST = "latest"


class FontGroup(Enum):
    SOLID = "solid"
    REGULAR = "regular"
    BRANDS = "brands"


def enable(
    source: str,
    group: FontGroup = FontGroup.REGULAR,
    force: bool = False,
    font_folder=Path.getcwd() / "fonts",
):
    """
    All-in-one function to deploy free Font Awesome fonts.
    It fetches the font file with the version from the given Font Awesome kit script file,
    converts the css and registers it as primary font.
    :param source: the script for your font kit from https://fontawesome.com/kits
        OR a Font Awesome version
        OR `~kivysome.LATEST`
    :param group: a valid ::class:`~kivysome.FontGroup`
    :param force: to ignore the existing files and refetch them by force
    :param font_folder: to folder to put the font files, "fonts" is used as default
    """
    font_folder = Path(font_folder)
    font_folder.mkdir_p()

    if semver.VersionInfo.isvalid(source) or source == LATEST:
        version = source
    else:
        pool_manager = urllib3.PoolManager()
        try:
            result = pool_manager.request("GET", source)
        except MaxRetryError as e:
            raise ValueError("the given link could not be accessed") from e
        if not result.status == 200:
            raise ValueError("the given link did not return a correct status code")
        content = result.data.decode("utf-8")

        if '"license":"free"' not in content:
            raise ValueError(
                "the given link is not referencing a free version of a Font Awesome kit"
            )
        version_match = _VERSION_MATCHER.match(content)
        if version_match is None:
            raise ValueError("no version could be parsed from the given link")

        version = version_match.group(1)

    if version == LATEST:
        repo = "FortAwesome/Font-Awesome"
        version = lastversion.latest(repo, "version")
    elif not semver.VersionInfo.isvalid(version):
        raise ValueError(f"matched version {version} is no valid semantic version")

    download_link = f"https://use.fontawesome.com/releases/v{version}/fontawesome-free-{version}-web.zip"

    ttf_local = font_folder / f"fa-{group.value}-{version}.ttf"
    css_local = ttf_local.with_suffix(".css")
    fontd_local = ttf_local.with_suffix(".fontd")

    if force or not ttf_local.isfile() or not css_local.isfile():

        try:
            with RemoteZip(download_link) as remote_zip:
                ttf_file: str = ""
                css_file: str = ""
                for file_name in remote_zip.namelist():
                    if group.value in file_name and file_name.endswith(".ttf"):
                        ttf_file = file_name
                    elif file_name.endswith("all.css"):
                        css_file = file_name

                if not ttf_file or not css_file:
                    raise ValueError(
                        f"the required font files could not be found in the remote zip file {download_link}"
                    )

                if ttf_local.exists():
                    if not remote_zip.getinfo(ttf_file).file_size == ttf_local.size:
                        ttf_local.remove()
                if css_local.exists():
                    if not remote_zip.getinfo(css_file).file_size == css_local.size:
                        css_local.remove()

                if not ttf_local.exists():
                    target_path = remote_zip.extract(ttf_file, font_folder)
                    Path(target_path).move(ttf_local)
                if not css_local.exists():
                    target_path = remote_zip.extract(css_file, font_folder)
                    Path(target_path).move(css_local)

                for tmp_folder in font_folder.dirs(f"*{version}*"):
                    if len(list(tmp_folder.walkfiles())) == 0:
                        tmp_folder.rmtree()
        except RemoteIOError as e:
            raise ValueError(
                f"the referenced Font Awesome version {version} could not be downloaded from GitHub"
            ) from e

    if force or not fontd_local.isfile():
        create_fontdict_file(css_local, fontd_local)

    register(f"fontawesome-{group}", ttf_local, fontd_local)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kivysome']

package_data = \
{'': ['*']}

install_requires = \
['lastversion==1.2.0',
 'mutapath>=0.16.0,<0.17.0',
 'remotezip>=0.9.2,<0.10.0',
 'semver>=2.10.2,<3.0.0',
 'urllib3>=1.25.7,<2.0.0']

setup_kwargs = {
    'name': 'kivysome',
    'version': '0.2.1',
    'description': 'Font Awesome 5 Icons for Kivy',
    'long_description': '# kivysome\n\n[![GitHub Workflow Status (branch)](https://img.shields.io/github/workflow/status/matfax/kivysome/build/master?style=for-the-badge)](https://github.com/matfax/kivysome/actions)\n[![Codecov](https://img.shields.io/codecov/c/github/matfax/kivysome?style=for-the-badge)](https://codecov.io/gh/matfax/kivysome)\n[![Dependabot Status](https://img.shields.io/badge/dependabot-enabled-blue?style=for-the-badge&logo=dependabot&color=0366d6)](https://github.com/matfax/kivysome/network/updates)\n[![Libraries.io dependency status for latest release](https://img.shields.io/librariesio/release/pypi/kivysome?style=for-the-badge)](https://libraries.io/pypi/kivysome)\n[![CodeFactor](https://www.codefactor.io/repository/github/matfax/kivysome/badge?style=for-the-badge)](https://www.codefactor.io/repository/github/matfax/kivysome)\n[![security: bandit](https://img.shields.io/badge/security-bandit-purple.svg?style=for-the-badge)](https://github.com/PyCQA/bandit)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg?style=for-the-badge)](https://github.com/psf/black)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/kivysome?style=for-the-badge)](https://pypi.org/project/kivysome/)\n[![PyPI](https://img.shields.io/pypi/v/kivysome?color=%2339A7A6&style=for-the-badge)](https://pypi.org/project/kivysome/)\n[![PyPI - Downloads](https://img.shields.io/pypi/dm/kivysome?color=ff69b4&style=for-the-badge)](https://pypistats.org/packages/kivysome)\n[![GitHub License](https://img.shields.io/github/license/matfax/kivysome.svg?style=for-the-badge)](https://github.com/matfax/kivysome/blob/master/LICENSE)\n[![GitHub last commit](https://img.shields.io/github/last-commit/matfax/kivysome?color=9cf&style=for-the-badge)](https://github.com/matfax/kivysome/commits/master)\n\nFont Awesome 5 Icons for Kivy\n\n## Usage\n\n### Enable it\n\n#### Using a version\n\nThis will only work for free versions of Font Awesome.\n\n```python\nimport kivysome \nkivysome.enable(kivysome.LATEST, group=kivysome.FontGroup.REGULAR)\n```\n\n#### Using a kit\n\nThis might be extended to commercial versions of Font Awesome on demand.\n\n##### 1. Generate your kit\n\nGo to [Font Awesome](https://fontawesome.com/kits) and generate your kit there.\nThe specified version is respected.\nFor the moment, only free licenses are supported. \n\n##### 2. Enable it\n\nIn your main.py register your font:\n\n```python\nimport kivysome \nkivysome.enable("https://kit.fontawesome.com/{YOURCODE}.js", group=kivysome.FontGroup.SOLID)\n```\n\n### 3. Use it\n\nIn your `.kv` file or string, reference the short Font Awesome (i.e., without `fa-` prefix) as you can copy them from their website.\n\n```yaml\n#: import icon kivysome.icon\nButton:\n    markup: True # Always turn markup on\n    text: "%s Comment" % icon(\'comment\', 24)\n```\n\n## Caching\n\nKivysome will cache the files in the font folder and not redownload them from GitHub.\nIf a kit is given, however, the kit version will have to be fetched from Font Awesome on every execution.\nIf `kivysome.LATEST` is given, the latest version will also have to be determined from GitHub\'s servers.\nFont packs will only be downloaded if new versions are published, then.\nTo completely avoid regular server access, a pinned version will have to be given.\nThe initial download can also be circumvented by downloading it a single time on the developer\'s machine and publishing\nthe downloaded `fonts` folder with the project. The folder should then contain a `.css`, `.fontd`, and a `.ttf` file\nmatching the pinned Font Awesome version and font group.\n\nCheck the `examples` folder for more insight.\n',
    'author': 'matfax',
    'author_email': 'matthias.fax@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/matfax/kivysome',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import json
import re
from mutapath import Path

_register = dict()


def register(name, ttf_fname, fontd_fname):
    """Register an Iconfont
    :param name: font name identifier.
    :param ttf_fname: ttf filename (path)
    :param fontd_fname: fontdic filename. (See create_fontdic)
    """
    font_file = Path(fontd_fname)
    fontd = json.loads(font_file.text)
    _register[name] = ttf_fname, fontd_fname, fontd


def icon(code, size=None, color=None, font_name=None):
    """Gets an icon from iconfont.
    :param code: Icon codename (ex: 'icon-name')
    :param size: Icon size
    :param color: Icon color
    :param font_name: Registered font name. If None first one is used.
    :returns: icon text (with markups)
    """
    font = next(iter(_register.keys())) if font_name is None else font_name
    font_data = _register[font]
    s = "[font=%s]%s[/font]" % (font_data[0], chr(font_data[2][code]))
    if size is not None:
        s = "[size=%s]%s[/size]" % (size, s)
    if color is not None:
        s = "[color=%s]%s[/color]" % (color, s)

    return s


def create_fontdict_file(css_fname, output_fname=""):
    """Creates a font dictionary file. Basically creates a dictionary filled
    with icon_code: unicode_value entries
    obtained from a CSS file.
    :param css_fname: CSS filename where font's rules are declared.
    :param output_fname: Fontd file destination
    """
    css_file = Path(css_fname)
    if output_fname:
        output_file = Path(output_fname)
    else:
        output_file = css_file.with_suffix(".fontd")

    res = _parse(css_file.read_text("utf-8"))
    output_file.write_text(json.dumps(res))

    return res


def _parse(data):
    # find start index where icons rules start
    pat_start = re.compile("}.+content:", re.DOTALL)
    rules_start = [x for x in re.finditer(pat_start, data)][0].start()
    data = data[rules_start:]  # crop data
    data = data.replace("\\", "0x")  # replace unicodes
    data = data.replace("'", '"')  # replace quotes
    # iterate rule indices and extract value
    pat_keys = re.compile("[a-zA-Z0-9_-]+:before")
    res = dict()
    for i in re.finditer(pat_keys, data):
        start = i.start()
        end = data.find("}", start)
        key = i.group().replace(":before", "")
        try:
            value = int(data[start:end].split('"')[1], 0)
        except (IndexError, ValueError):
            continue
        if key.startswith("fa-"):
            _, key = key.split("-", 1)
        res[key] = value
    return res
